"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.wrapIdentifier = void 0;
const cassandra = __importStar(require("cassandra-driver"));
const logger_1 = __importDefault(require("../logger"));
const utils_1 = require("../utils");
const abstract_adapter_1 = require("./abstract_adapter");
const logger = logger_1.default('db:clients:cassandra');
/**
 * To keep compatibility with the other clients we treat keyspaces as database.
 */
class CassandraAdapter extends abstract_adapter_1.AbstractAdapter {
    constructor(server, database) {
        super(server, database);
        const dbConfig = this.configDatabase();
        logger().debug('creating database client %j', dbConfig);
        this.client = new cassandra.Client(dbConfig);
    }
    configDatabase() {
        const config = {
            contactPoints: [this.server.config.host],
            protocolOptions: {
                port: this.server.config.port,
            },
            keyspace: this.database.database,
        };
        if (this.server.sshTunnel) {
            config.contactPoints = [this.server.config.localHost];
            config.protocolOptions.port = this.server.config.localPort;
        }
        if (this.server.config.ssl) {
            // TODO: sslOptions
        }
        // client authentication
        if (this.server.config.user && this.server.config.password) {
            const user = this.server.config.user;
            const password = this.server.config.password;
            const authProviderInfo = new cassandra.auth.PlainTextAuthProvider(user, password);
            config.authProvider = authProviderInfo;
        }
        return config;
    }
    connect() {
        logger().debug('connecting');
        return new Promise((resolve, reject) => {
            this.client.connect((err) => {
                if (err) {
                    this.client.shutdown();
                    return reject(err);
                }
                this.version = {
                    name: 'Cassandra',
                    version: this.client.getState().getConnectedHosts()[0].cassandraVersion,
                    string: `Cassandra ${this.client.getState().getConnectedHosts()[0].cassandraVersion}`,
                };
                logger().debug('connected');
                resolve();
            });
        });
    }
    async disconnect() {
        return new Promise((resolve, reject) => {
            this.client.shutdown((err) => {
                if (err) {
                    return reject(err);
                }
                resolve();
            });
        });
    }
    listDatabases() {
        return new Promise((resolve) => {
            resolve(Object.keys(this.client.metadata.keyspaces));
        });
    }
    listTables() {
        return new Promise((resolve, reject) => {
            let sql;
            if (this.version.version[0] === '2') {
                sql = `
          SELECT columnfamily_name as name
          FROM system.schema_columnfamilies
          WHERE keyspace_name = ?
        `;
            }
            else {
                sql = `
          SELECT table_name as name
          FROM system_schema.tables
          WHERE keyspace_name = ?
        `;
            }
            const params = [this.database.database];
            this.client.execute(sql, params, (err, data) => {
                if (err)
                    return reject(err);
                resolve(data.rows.map((row) => ({ name: row.name })));
            });
        });
    }
    listTableColumns(table) {
        const cassandra2 = this.version.version[0] === '2';
        return new Promise((resolve, reject) => {
            let sql;
            if (cassandra2) {
                sql = `
          SELECT type as position, column_name, validator as type
          FROM system.schema_columns
          WHERE keyspace_name = ?
            AND columnfamily_name = ?
        `;
            }
            else {
                sql = `
          SELECT position, column_name, type
          FROM system_schema.columns
          WHERE keyspace_name = ?
            AND table_name = ?
        `;
            }
            const params = [this.database.database, table];
            this.client.execute(sql, params, (err, data) => {
                if (err)
                    return reject(err);
                resolve(data.rows
                    // force pks be placed at the results beginning
                    .sort((a, b) => {
                    if (cassandra2) {
                        return +(a.position > b.position) || -(a.position < b.position);
                    }
                    return b.position - a.position;
                })
                    .map((row) => {
                    const rowType = cassandra2
                        ? mapLegacyDataTypes(row.type)
                        : row.type;
                    return {
                        columnName: row.column_name,
                        dataType: rowType,
                    };
                }));
            });
        });
    }
    getTableKeys(table) {
        return new Promise((resolve, reject) => {
            if (this.database.database === undefined) {
                return [];
            }
            this.client.metadata.getTable(this.database.database, table, (err, tableInfo) => {
                if (err) {
                    return reject(err);
                }
                resolve(tableInfo.partitionKeys.map((key) => ({
                    constraintName: null,
                    columnName: key.name,
                    referencedTable: null,
                    keyType: 'PRIMARY KEY',
                })));
            });
        });
    }
    async truncateAllTables() {
        const result = await this.listTables();
        const tables = result.map((table) => table.name);
        const promises = tables.map((t) => {
            if (this.database.database === undefined) {
                return;
            }
            const truncateSQL = `
        TRUNCATE TABLE ${this.wrapIdentifier(this.database.database)}.${this.wrapIdentifier(t)};
      `;
            return this.executeQuery(truncateSQL);
        });
        await Promise.all(promises);
    }
    query() {
        throw new Error('"query" function is not implementd by cassandra client.');
    }
    executeQuery(queryText) {
        const commands = utils_1.identifyCommands(queryText).map((item) => item.type);
        return new Promise((resolve, reject) => {
            this.client.execute(queryText, (err, data) => {
                if (err)
                    return reject(err);
                resolve([parseRowQueryResult(data, commands[0])]);
            });
        });
    }
    wrapIdentifier(value) {
        return wrapIdentifier(value);
    }
}
exports.default = CassandraAdapter;
function wrapIdentifier(value) {
    if (value === '*')
        return value;
    const matched = /(.*?)(\[[0-9]\])/.exec(value);
    if (matched)
        return wrapIdentifier(matched[1]) + matched[2];
    return `"${value.replace(/"/g, '""')}"`;
}
exports.wrapIdentifier = wrapIdentifier;
/**
 * The system schema of Casandra 2.x does not have data type, but only validator
 * classes. To make the behavior consistent with v3.x, we try to deduce the
 * correponding CQL data type using the validator name.
 */
function mapLegacyDataTypes(validator) {
    const type = validator.split('.').pop();
    switch (type) {
        case 'Int32Type':
        case 'LongType':
            return 'int';
        case 'UTF8Type':
            return 'text';
        case 'TimestampType':
        case 'DateType':
            return 'timestamp';
        case 'DoubleType':
            return 'double';
        case 'FloatType':
            return 'float';
        case 'UUIDType':
            return 'uuid';
        case 'CounterColumnType':
            return 'counter';
        default:
            logger().debug('validator %s is not yet mapped!', validator);
            return type;
    }
}
function parseRowQueryResult(data, command) {
    // Fallback in case the identifier could not reconize the command
    const isSelect = command ? command === 'SELECT' : Array.isArray(data.rows);
    return {
        command: command || (isSelect && 'SELECT'),
        rows: data.rows || [],
        fields: data.columns || [],
        rowCount: isSelect ? data.rowLength || 0 : undefined,
        affectedRows: !isSelect && !isNaN(data.rowLength) ? data.rowLength : undefined,
    };
}
