"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.wrapIdentifier = void 0;
const mysql2_1 = __importDefault(require("mysql2"));
const sql_query_identifier_1 = require("sql-query-identifier");
const logger_1 = __importDefault(require("../logger"));
const utils_1 = require("../utils");
const abstract_adapter_1 = require("./abstract_adapter");
const logger = logger_1.default('db:clients:mysql');
const mysqlErrors = {
    EMPTY_QUERY: 'ER_EMPTY_QUERY',
    CONNECTION_LOST: 'PROTOCOL_CONNECTION_LOST',
};
class MysqlAdapter extends abstract_adapter_1.AbstractAdapter {
    constructor(server, database) {
        super(server, database);
        const dbConfig = this.configDatabase();
        logger().debug('create adapter for mysql with config %j', dbConfig);
        this.conn = {
            pool: mysql2_1.default.createPool(dbConfig),
        };
    }
    configDatabase() {
        const config = {
            host: this.server.config.host,
            port: this.server.config.port,
            user: this.server.config.user,
            password: this.server.config.password,
            database: this.database.database,
            multipleStatements: true,
            dateStrings: true,
            supportBigNumbers: true,
            bigNumberStrings: true,
        };
        if (this.server.sshTunnel) {
            config.host = this.server.config.localHost;
            config.port = this.server.config.localPort;
        }
        if (this.server.config.ssl) {
            config.ssl = {
                // It is not the best recommend way to use SSL with node-mysql
                // https://github.com/felixge/node-mysql#ssl-options
                // But this way we have compatibility with all clients.
                rejectUnauthorized: false,
            };
        }
        return config;
    }
    async connect() {
        logger().debug('connecting');
        const versionInfo = (await this.driverExecuteQuery({
            query: "SHOW VARIABLES WHERE variable_name='version' OR variable_name='version_comment';",
        })).data;
        let version = '';
        let versionComment = '';
        for (let i = 0; i < versionInfo.length; i++) {
            const item = versionInfo[i];
            if (item.Variable_name === 'version') {
                version = item.Value;
            }
            else if (item.Variable_name === 'version_comment') {
                versionComment = item.Value;
            }
        }
        this.version = {
            name: versionComment,
            version: version.split('-')[0],
            string: `${versionComment} ${version}`,
        };
        // normalize the name as depending on where the server is installed from, it
        // could be just "MySQL" or "MariaDB", or it could be a longer string like
        // "mariadb.org binary distribution"
        const lowerComment = versionComment.toLowerCase();
        if (lowerComment.includes('mysql')) {
            this.version.name = 'MySQL';
        }
        else if (lowerComment.includes('mariadb')) {
            this.version.name = 'MariaDB';
        }
        else if (lowerComment.includes('percona')) {
            this.version.name = 'Percona';
        }
        logger().debug('connected');
    }
    disconnect() {
        return new Promise((resolve, reject) => {
            this.conn.pool.end((err) => {
                if (err) {
                    return reject(err);
                }
                resolve();
            });
        });
    }
    query(queryText) {
        let pid = null;
        let canceling = false;
        const cancelable = utils_1.createCancelablePromise();
        return {
            execute: () => {
                return this.runWithConnection(async (connection) => {
                    const pidResult = await this.driverExecuteQuery({
                        query: 'SELECT connection_id() AS pid',
                    }, connection);
                    pid = pidResult.data[0].pid;
                    try {
                        const data = await Promise.race([
                            cancelable.wait(),
                            this.executeQuery(queryText, connection),
                        ]);
                        pid = null;
                        return data;
                    }
                    catch (err) {
                        if (canceling && err.code === mysqlErrors.CONNECTION_LOST) {
                            canceling = false;
                            err.sqlectronError = 'CANCELED_BY_USER';
                        }
                        throw err;
                    }
                    finally {
                        cancelable.discard();
                    }
                });
            },
            cancel: async () => {
                if (!pid) {
                    throw new Error('Query not ready to be canceled');
                }
                canceling = true;
                try {
                    await this.driverExecuteQuery({
                        query: `kill ${pid};`,
                    });
                    cancelable.cancel();
                }
                catch (err) {
                    canceling = false;
                    throw err;
                }
            },
        };
    }
    async executeQuery(queryText, connection) {
        const { data, fields } = await this.driverExecuteQuery({ query: queryText }, connection);
        if (!data || (Array.isArray(data) && data.length === 0 && fields.length === 0)) {
            return [];
        }
        const commands = identifyCommands(queryText).map((item) => item.type);
        if (!isMultipleQuery(fields)) {
            return [
                parseRowQueryResult(data, fields, commands[0]),
            ];
        }
        return data.map((_, idx) => {
            return parseRowQueryResult(data[idx], fields[idx], commands[idx]);
        });
    }
    async listTables() {
        const sql = `
      SELECT table_name as name
      FROM information_schema.tables
      WHERE table_schema = database()
      AND table_type NOT LIKE '%VIEW%'
      ORDER BY table_name
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data;
    }
    async listViews() {
        const sql = `
      SELECT table_name as name
      FROM information_schema.views
      WHERE table_schema = database()
      ORDER BY table_name
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data;
    }
    async listRoutines() {
        const sql = `
      SELECT routine_name as 'routine_name', routine_type as 'routine_type'
      FROM information_schema.routines
      WHERE routine_schema = database()
      ORDER BY routine_name
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => ({
            routineName: row.routine_name,
            routineType: row.routine_type,
        }));
    }
    async listTableColumns(table) {
        const sql = `
      SELECT column_name AS 'column_name', data_type AS 'data_type'
      FROM information_schema.columns
      WHERE table_schema = database()
      AND table_name = ?
      ORDER BY ordinal_position
    `;
        const params = [table];
        const { data } = await this.driverExecuteQuery({ query: sql, params });
        return data.map((row) => ({
            columnName: row.column_name,
            dataType: row.data_type,
        }));
    }
    async listTableTriggers(table) {
        const sql = `
      SELECT trigger_name as 'trigger_name'
      FROM information_schema.triggers
      WHERE event_object_schema = database()
      AND event_object_table = ?
    `;
        const params = [table];
        const { data } = await this.driverExecuteQuery({ query: sql, params });
        return data.map((row) => row.trigger_name);
    }
    async listTableIndexes(table) {
        const sql = 'SHOW INDEX FROM ?? FROM ??';
        const params = [table, this.database.database];
        const { data } = await this.driverExecuteQuery({ query: sql, params });
        return data.map((row) => row.Key_name);
    }
    async listDatabases(filter) {
        const sql = 'show databases';
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data
            .filter((item) => filterDatabase(item, filter, 'Database'))
            .map((row) => row.Database);
    }
    async getTableReferences(table) {
        const sql = `
      SELECT referenced_table_name as 'referenced_table_name'
      FROM information_schema.key_column_usage
      WHERE referenced_table_name IS NOT NULL
      AND table_schema = database()
      AND table_name = ?
    `;
        const params = [table];
        const { data } = await this.driverExecuteQuery({ query: sql, params });
        return data.map((row) => row.referenced_table_name);
    }
    async getTableKeys(table) {
        const sql = `
      SELECT
        constraint_name as 'constraint_name',
        column_name as 'column_name',
        referenced_table_name as 'referenced_table_name',
        CASE WHEN (referenced_table_name IS NOT NULL) THEN 'FOREIGN'
        ELSE constraint_name
        END as key_type
      FROM information_schema.key_column_usage
      WHERE table_schema = database()
      AND table_name = ?
      AND ((referenced_table_name IS NOT NULL) OR constraint_name LIKE '%PRIMARY%')
    `;
        const params = [table];
        const { data } = await this.driverExecuteQuery({ query: sql, params });
        return data.map((row) => ({
            constraintName: `${row.constraint_name} KEY`,
            columnName: row.column_name,
            referencedTable: row.referenced_table_name,
            keyType: `${row.key_type} KEY`,
        }));
    }
    async getTableCreateScript(table) {
        const sql = `SHOW CREATE TABLE ${table}`;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => utils_1.appendSemiColon(row['Create Table']));
    }
    async getViewCreateScript(view) {
        const sql = `SHOW CREATE VIEW ${view}`;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => utils_1.appendSemiColon(row['Create View']));
    }
    async getRoutineCreateScript(routine, type) {
        const sql = `SHOW CREATE ${type.toUpperCase()} ${routine}`;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => utils_1.appendSemiColon(row[`Create ${type}`]));
    }
    async getSchema(connection) {
        const sql = "SELECT database() AS 'schema'";
        const result = await this.driverExecuteQuery({ query: sql }, connection);
        return result.data[0].schema;
    }
    async truncateAllTables() {
        await this.runWithConnection(async (connection) => {
            const schema = await this.getSchema(connection);
            const sql = `
        SELECT table_name as 'table_name'
        FROM information_schema.tables
        WHERE table_schema = '${schema}'
        AND table_type NOT LIKE '%VIEW%'
      `;
            const result = await this.driverExecuteQuery({ query: sql }, connection);
            const data = result.data;
            const truncateAll = data
                .map((row) => `
        SET FOREIGN_KEY_CHECKS = 0;
        TRUNCATE TABLE ${this.wrapIdentifier(schema)}.${this.wrapIdentifier(row.table_name)};
        SET FOREIGN_KEY_CHECKS = 1;
      `)
                .join('');
            return this.driverExecuteQuery({ query: truncateAll }, connection);
        });
    }
    async driverExecuteQuery(queryArgs, connection) {
        const runQuery = (connection) => {
            return new Promise((resolve, reject) => {
                connection.query(queryArgs.query, queryArgs.params, (err, data, fields) => {
                    if (err && err.code === mysqlErrors.EMPTY_QUERY) {
                        return resolve({ data: [], fields: [] });
                    }
                    if (err) {
                        return reject(getRealError(connection, err));
                    }
                    resolve({
                        data: data,
                        fields,
                    });
                });
            });
        };
        return connection ? runQuery(connection) : this.runWithConnection(runQuery);
    }
    runWithConnection(run) {
        let rejected = false;
        return new Promise((resolve, reject) => {
            const rejectErr = (err) => {
                if (!rejected) {
                    rejected = true;
                    reject(err);
                }
            };
            this.conn.pool.getConnection(async (errPool, connection) => {
                if (errPool) {
                    rejectErr(errPool);
                    return;
                }
                connection.on('error', (error) => {
                    // it will be handled later in the next query execution
                    logger().error('Connection fatal error %j', error);
                });
                try {
                    resolve(await run(connection));
                }
                catch (err) {
                    rejectErr(err);
                }
                finally {
                    connection.release();
                }
            });
        });
    }
    wrapIdentifier(value) {
        return wrapIdentifier(value);
    }
}
exports.default = MysqlAdapter;
function wrapIdentifier(value) {
    return value !== '*' ? `\`${value.replace(/`/g, '``')}\`` : '*';
}
exports.wrapIdentifier = wrapIdentifier;
function identifyCommands(queryText) {
    try {
        return sql_query_identifier_1.identify(queryText, { strict: false });
    }
    catch (err) {
        return [];
    }
}
function getRealError(conn, err) {
    /* eslint no-underscore-dangle:0 */
    if (conn && (conn._fatalError || conn._protocolError)) {
        return conn._fatalError || conn._protocolError;
    }
    return err;
}
function parseRowQueryResult(data, fields, command) {
    // Fallback in case the identifier could not reconize the command
    const isSelect = Array.isArray(data);
    const dataArray = isSelect ? data || [] : [];
    const dataHeader = !isSelect ? data : { affectedRows: undefined };
    return {
        command: command || (isSelect ? 'SELECT' : 'UNKNOWN'),
        rows: dataArray,
        fields: fields || [],
        rowCount: isSelect ? dataArray.length : undefined,
        affectedRows: dataHeader.affectedRows,
    };
}
function isMultipleQuery(fields) {
    if (!fields) {
        return false;
    }
    if (!fields.length) {
        return false;
    }
    return Array.isArray(fields[0]) || fields[0] === undefined;
}
function filterDatabase(item, { database } = {}, databaseField) {
    if (!database) {
        return true;
    }
    const value = item[databaseField];
    if (typeof database === 'string') {
        return database === value;
    }
    const { only, ignore } = database;
    if (only && only.length && !only.includes(value)) {
        return false;
    }
    if (ignore && ignore.length && ignore.includes(value)) {
        return false;
    }
    return true;
}
