"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.wrapIdentifier = void 0;
const sqlite3_1 = __importDefault(require("sqlite3"));
const sql_query_identifier_1 = require("sql-query-identifier");
const logger_1 = __importDefault(require("../logger"));
const utils_1 = require("../utils");
const _1 = require("./");
const abstract_adapter_1 = require("./abstract_adapter");
const logger = logger_1.default('db:clients:sqlite');
const sqliteErrors = {
    CANCELED: 'SQLITE_INTERRUPT',
};
class SqliteAdapter extends abstract_adapter_1.AbstractAdapter {
    constructor(server, database) {
        super(server, database);
        const dbConfig = this.configDatabase();
        logger().debug('create adapter for sqlite3 with config %j', dbConfig);
        this.conn = { dbConfig };
    }
    configDatabase() {
        return {
            database: this.database.database ||
                _1.ADAPTERS.find((adapter) => adapter.key === 'sqlite').defaultDatabase,
        };
    }
    // SQLite does not have connection poll. So we open and close connections
    // for every query request. This allows multiple request at same time by
    // using a different thread for each connection.
    // This may cause connection limit problem. So we may have to change this at some point.
    async connect() {
        var _a;
        logger().debug('connecting');
        const result = (await this.driverExecuteQuery({ query: 'SELECT sqlite_version() as version' }));
        if (!result.data || ((_a = result.data) === null || _a === void 0 ? void 0 : _a.length) === 0) {
            throw new Error('Failed to fetch version information');
        }
        const version = result.data[0].version;
        this.version = {
            name: 'SQLite',
            version,
            string: `SQLite ${version}`,
        };
        logger().debug('connected');
    }
    query(queryText) {
        let queryConnection = null;
        return {
            execute: () => {
                return this.runWithConnection(async (connection) => {
                    try {
                        queryConnection = connection;
                        const result = await this.executeQuery(queryText, connection);
                        return result;
                    }
                    catch (err) {
                        if (err.code === sqliteErrors.CANCELED) {
                            err.sqlectronError = 'CANCELED_BY_USER';
                        }
                        throw err;
                    }
                });
            },
            cancel: () => {
                if (!queryConnection) {
                    throw new Error('Query not ready to be canceled');
                }
                queryConnection.interrupt();
            },
        };
    }
    async executeQuery(queryText, connection) {
        const result = await this.driverExecuteQuery({ query: queryText, multiple: true }, connection);
        return result.map((value) => {
            return parseRowQueryResult(value);
        });
    }
    async listTables(filter, connection) {
        const sql = `
      SELECT name
      FROM sqlite_master
      WHERE type='table'
      ORDER BY name
    `;
        const { data } = await this.driverExecuteQuery({ query: sql }, connection);
        return data;
    }
    async listViews() {
        const sql = `
      SELECT name
      FROM sqlite_master
      WHERE type = 'view'
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data;
    }
    async listTableColumns(table) {
        const sql = `PRAGMA table_info('${table}')`;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => ({
            columnName: row.name,
            dataType: row.type,
        }));
    }
    async listTableTriggers(table) {
        const sql = `
      SELECT name
      FROM sqlite_master
      WHERE type = 'trigger'
        AND tbl_name = '${table}'
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => row.name);
    }
    async listTableIndexes(table) {
        const sql = `PRAGMA INDEX_LIST('${table}')`;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => row.name);
    }
    async listDatabases() {
        const sql = 'PRAGMA database_list;';
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => row.file || ':memory:');
    }
    async getTableCreateScript(table) {
        const sql = `
      SELECT sql
      FROM sqlite_master
      WHERE name = '${table}';
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => utils_1.appendSemiColon(row.sql));
    }
    async getViewCreateScript(view) {
        const sql = `
      SELECT sql
      FROM sqlite_master
      WHERE name = '${view}';
    `;
        const { data } = await this.driverExecuteQuery({ query: sql });
        return data.map((row) => utils_1.appendSemiColon(row.sql));
    }
    async truncateAllTables() {
        await this.runWithConnection(async (connection) => {
            const tables = await this.listTables(null, connection);
            const truncateAll = tables
                .map((table) => `
        DELETE FROM ${table.name};
      `)
                .join('');
            // TODO: Check if sqlite_sequence exists then execute:
            // DELETE FROM sqlite_sequence WHERE name='${table}';
            await this.driverExecuteQuery({ query: truncateAll }, connection);
        });
    }
    async driverExecuteQuery(queryArgs, connection) {
        const runQuery = (connection, { executionType, text }) => new Promise((resolve, reject) => {
            const method = resolveExecutionType(executionType);
            connection[method](text, queryArgs.params, function (err, data) {
                if (err) {
                    return reject(err);
                }
                return resolve({
                    data,
                    lastID: this.lastID,
                    changes: this.changes, // eslint-disable-line @typescript-eslint/no-unnecessary-type-assertion
                });
            });
        });
        const identifyStatementsRunQuery = async (connection) => {
            const statements = identifyCommands(queryArgs.query);
            const results = await Promise.all(statements.map(async (statement) => {
                const result = await runQuery(connection, statement);
                return {
                    ...result,
                    statement,
                };
            }));
            return queryArgs.multiple ? results : results[0];
        };
        return connection
            ? identifyStatementsRunQuery(connection)
            : this.runWithConnection(identifyStatementsRunQuery);
    }
    runWithConnection(run) {
        return new Promise((resolve, reject) => {
            const db = new sqlite3_1.default.Database(this.conn.dbConfig.database, (err) => {
                if (err) {
                    reject(err);
                    return;
                }
                db.serialize();
                run(db)
                    .then((results) => {
                    resolve(results);
                })
                    .catch((runErr) => {
                    reject(runErr);
                })
                    .finally(() => {
                    db.close();
                });
            });
        });
    }
    wrapIdentifier(value) {
        return wrapIdentifier(value);
    }
}
exports.default = SqliteAdapter;
function wrapIdentifier(value) {
    if (value === '*')
        return value;
    const matched = /(.*?)(\[[0-9]\])/.exec(value);
    if (matched)
        return wrapIdentifier(matched[1]) + matched[2];
    return `"${value.replace(/"/g, '""')}"`;
}
exports.wrapIdentifier = wrapIdentifier;
function parseRowQueryResult({ data, statement, changes, }) {
    // Fallback in case the identifier could not reconize the command
    const isSelect = Array.isArray(data);
    const rows = data || [];
    return {
        command: statement.type || (isSelect && 'SELECT'),
        rows,
        fields: Object.keys(rows[0] || {}).map((name) => ({ name })),
        rowCount: data && data.length,
        affectedRows: changes || 0,
    };
}
function identifyCommands(queryText) {
    try {
        return sql_query_identifier_1.identify(queryText, { strict: false });
    }
    catch (err) {
        return [];
    }
}
function resolveExecutionType(executioType) {
    switch (executioType) {
        case 'MODIFICATION':
            return 'run';
        default:
            return 'all';
    }
}
