"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.wrapIdentifier = void 0;
const mssql_1 = require("mssql");
const filters_1 = require("../filters");
const logger_1 = __importDefault(require("../logger"));
const utils_1 = require("../utils");
const abstract_adapter_1 = require("./abstract_adapter");
const logger = logger_1.default('db:clients:sqlserver');
const mmsqlErrors = {
    CANCELED: 'ECANCEL',
};
class SqlServerAdapter extends abstract_adapter_1.AbstractAdapter {
    constructor(server, database) {
        super(server, database);
        const dbConfig = this.configDatabase();
        logger().debug('create driver client for mmsql with config %j', dbConfig);
        this.conn = { dbConfig };
    }
    configDatabase() {
        const config = {
            user: this.server.config.user,
            password: this.server.config.password,
            server: this.server.config.host,
            database: this.database.database,
            port: this.server.config.port,
            requestTimeout: Infinity,
            domain: this.server.config.domain,
            pool: {
                max: 5,
            },
            options: {
                encrypt: !!this.server.config.ssl,
                appName: this.server.config.applicationName || 'sqlectron',
                enableArithAbort: true,
            },
        };
        if (this.server.sshTunnel) {
            config.server = this.server.config.localHost;
            config.port = this.server.config.localPort;
        }
        return config;
    }
    async connect() {
        logger().debug('connecting');
        const version = (await this.driverExecuteSingleQuery({
            query: "SELECT @@version as 'version'",
        })).data[0].version;
        this.version = {
            name: 'SQL Server',
            version: /^Microsoft SQL Server ([0-9]{4})/.exec(version)[1],
            string: version,
        };
        logger().debug('connected');
    }
    async disconnect() {
        if (!this.conn.connection) {
            return;
        }
        return this.conn.connection.close();
    }
    async getSchema(connection) {
        const sql = "SELECT schema_name() AS 'schema'";
        const { data } = await this.driverExecuteSingleQuery({ query: sql }, connection);
        return data[0].schema;
    }
    async listDatabases(filter) {
        const databaseFilter = filters_1.buildDatabaseFilter(filter, 'name');
        const sql = `
      SELECT name
      FROM sys.databases
      ${databaseFilter ? `AND ${databaseFilter}` : ''}
      ORDER BY name
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => row.name);
    }
    async listSchemas(filter) {
        const schemaFilter = filters_1.buildSchemaFilter(filter);
        const sql = `
      SELECT schema_name
      FROM INFORMATION_SCHEMA.SCHEMATA
      ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
      ORDER BY schema_name
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => row.schema_name);
    }
    async listTables(filter) {
        const schemaFilter = filters_1.buildSchemaFilter(filter, 'table_schema');
        const sql = `
      SELECT
        table_schema,
        table_name
      FROM INFORMATION_SCHEMA.TABLES
      WHERE table_type NOT LIKE '%VIEW%'
      ${schemaFilter ? `AND ${schemaFilter}` : ''}
      ORDER BY table_schema, table_name
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((item) => ({
            schema: item.table_schema,
            name: item.table_name,
        }));
    }
    async listViews(filter) {
        const schemaFilter = filters_1.buildSchemaFilter(filter, 'table_schema');
        const sql = `
      SELECT
        table_schema,
        table_name
      FROM INFORMATION_SCHEMA.VIEWS
      ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
      ORDER BY table_schema, table_name
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((item) => ({
            schema: item.table_schema,
            name: item.table_name,
        }));
    }
    async listRoutines(filter) {
        const schemaFilter = filters_1.buildSchemaFilter(filter, 'routine_schema');
        const sql = `
      SELECT
        routine_schema,
        routine_name,
        routine_type
      FROM INFORMATION_SCHEMA.ROUTINES
      ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
      GROUP BY routine_schema, routine_name, routine_type
      ORDER BY routine_schema, routine_name
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => ({
            schema: row.routine_schema,
            routineName: row.routine_name,
            routineType: row.routine_type,
        }));
    }
    async listTableColumns(table) {
        const sql = `
      SELECT column_name, data_type
      FROM INFORMATION_SCHEMA.COLUMNS
      WHERE table_name = '${table}'
      ORDER BY ordinal_position
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => ({
            columnName: row.column_name,
            dataType: row.data_type,
        }));
    }
    async listTableTriggers(table) {
        // SQL Server does not have information_schema for triggers, so other way around
        // is using sp_helptrigger stored procedure to fetch triggers related to table
        const sql = `EXEC sp_helptrigger ${wrapIdentifier(table)}`;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => row.trigger_name);
    }
    async listTableIndexes(table) {
        // SQL Server does not have information_schema for indexes, so other way around
        // is using sp_helpindex stored procedure to fetch indexes related to table
        const sql = `EXEC sp_helpindex ${wrapIdentifier(table)}`;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => row.index_name);
    }
    async getTableReferences(table) {
        const sql = `
      SELECT OBJECT_NAME(referenced_object_id) referenced_table_name
      FROM sys.foreign_keys
      WHERE parent_object_id = OBJECT_ID('${table}')
    `;
        const { data } = await this.driverExecuteSingleQuery({
            query: sql,
        });
        return data.map((row) => row.referenced_table_name);
    }
    async getTableKeys(table) {
        const sql = `
      SELECT
        tc.constraint_name,
        kcu.column_name,
        CASE WHEN tc.constraint_type LIKE '%FOREIGN%' THEN OBJECT_NAME(sfk.referenced_object_id)
        ELSE NULL
        END AS referenced_table_name,
        tc.constraint_type
      FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS tc
      JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS kcu
        ON tc.constraint_name = kcu.constraint_name
      JOIN sys.foreign_keys as sfk
        ON sfk.parent_object_id = OBJECT_ID(tc.table_name)
      WHERE tc.table_name = '${table}'
      AND tc.constraint_type IN ('PRIMARY KEY', 'FOREIGN KEY')
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => ({
            constraintName: row.constraint_name,
            columnName: row.column_name,
            referencedTable: row.referenced_table_name,
            keyType: row.constraint_type,
        }));
    }
    getQuerySelectTop(table, limit) {
        return `SELECT TOP ${limit} * FROM ${this.wrapIdentifier(table)}`;
    }
    async getTableCreateScript(table) {
        // Reference http://stackoverflow.com/a/317864
        const sql = `
      SELECT  ('CREATE TABLE ' + so.name + ' (' +
        CHAR(13)+CHAR(10) + REPLACE(o.list, '&#x0D;', CHAR(13)) +
        ');' + CHAR(13)+CHAR(10) +
        CASE WHEN tc.constraint_name IS NULL THEN ''
             ELSE + CHAR(13)+CHAR(10) + 'ALTER TABLE ' + so.Name +
             ' ADD CONSTRAINT ' + tc.constraint_name  +
             ' PRIMARY KEY ' + '(' + LEFT(j.list, Len(j.list)-1) + ');'
        END) AS createtable
      FROM sysobjects so
      CROSS APPLY
        (SELECT
          '  ' + column_name + ' ' +
          data_type +
          CASE data_type
              WHEN 'sql_variant' THEN ''
              WHEN 'text' THEN ''
              WHEN 'ntext' THEN ''
              WHEN 'xml' THEN ''
              WHEN 'decimal' THEN '(' + cast(numeric_precision AS varchar) + ', '
                    + cast(numeric_scale AS varchar) + ')'
              ELSE coalesce('('+ CASE WHEN character_maximum_length = -1
                    THEN 'MAX'
                    ELSE cast(character_maximum_length AS varchar)
                  END + ')','')
            END + ' ' +
            CASE WHEN EXISTS (
              SELECT id FROM syscolumns
              WHERE object_name(id)=so.name
              AND name=column_name
              AND columnproperty(id,name,'IsIdentity') = 1
            ) THEN
              'IDENTITY(' +
              cast(ident_seed(so.name) AS varchar) + ',' +
              cast(ident_incr(so.name) AS varchar) + ')'
            ELSE ''
            END + ' ' +
             (CASE WHEN UPPER(IS_NULLABLE) = 'NO'
                   THEN 'NOT '
                   ELSE ''
            END ) + 'NULL' +
            CASE WHEN INFORMATION_SCHEMA.COLUMNS.column_default IS NOT NULL
                 THEN 'DEFAULT '+ INFORMATION_SCHEMA.COLUMNS.column_default
                 ELSE ''
            END + ',' + CHAR(13)+CHAR(10)
         FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name = so.name
         ORDER BY ordinal_position
         FOR XML PATH('')
      ) o (list)
      LEFT JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc
      ON  tc.table_name       = so.name
      AND tc.constraint_type  = 'PRIMARY KEY'
      CROSS APPLY
          (SELECT column_name + ', '
           FROM   INFORMATION_SCHEMA.KEY_COLUMN_USAGE kcu
           WHERE  kcu.constraint_name = tc.constraint_name
           ORDER BY ordinal_position
           FOR XML PATH('')
          ) j (list)
      WHERE   xtype = 'U'
      AND name    NOT IN ('dtproperties')
      AND so.name = '${table}'
    `;
        const { data } = await this.driverExecuteSingleQuery({ query: sql });
        return data.map((row) => row.createtable.replace(/\r\n/g, '\n'));
    }
    async getViewCreateScript(view) {
        const sql = `SELECT OBJECT_DEFINITION (OBJECT_ID('${view}')) AS ViewDefinition;`;
        const { data } = await this.driverExecuteSingleQuery({
            query: sql,
        });
        return data.map((row) => utils_1.appendSemiColon(row.ViewDefinition));
    }
    async getRoutineCreateScript(routine) {
        const sql = `
      SELECT routine_definition
      FROM INFORMATION_SCHEMA.ROUTINES
      WHERE routine_name = '${routine}'
    `;
        const { data } = await this.driverExecuteSingleQuery({
            query: sql,
        });
        return data.map((row) => utils_1.appendSemiColon(row.routine_definition));
    }
    async truncateAllTables() {
        await this.runWithConnection(async (connection) => {
            const schema = await this.getSchema(connection);
            const sql = `
        SELECT table_name
        FROM INFORMATION_SCHEMA.TABLES
        WHERE table_schema = '${schema}'
        AND table_type NOT LIKE '%VIEW%'
      `;
            const { data } = await this.driverExecuteSingleQuery({ query: sql }, connection);
            const disableForeignKeys = data
                .map((row) => `
        ALTER TABLE ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)} NOCHECK CONSTRAINT all;
      `)
                .join('');
            const truncateAll = data
                .map((row) => `
        DELETE FROM ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)};
        DBCC CHECKIDENT ('${schema}.${row.table_name}', RESEED, 0);
      `)
                .join('');
            const enableForeignKeys = data
                .map((row) => `
        ALTER TABLE ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)} WITH CHECK CHECK CONSTRAINT all;
      `)
                .join('');
            await this.driverExecuteQuery({
                query: disableForeignKeys + truncateAll + enableForeignKeys,
                multiple: true,
            }, connection);
        });
    }
    query(queryText) {
        let queryRequest = null;
        return {
            execute: () => {
                return this.runWithConnection(async (connection) => {
                    const request = connection.request();
                    request.multiple = true;
                    try {
                        const promiseQuery = request.query(queryText);
                        queryRequest = request;
                        const result = await promiseQuery;
                        const data = request.multiple ? result.recordsets : result.recordset;
                        const affectedRows = result.rowsAffected
                            ? result.rowsAffected.reduce((a, b) => a + b, 0)
                            : undefined;
                        const commands = utils_1.identifyCommands(queryText).map((item) => item.type);
                        // Executing only non select queries will not return results.
                        // So we "fake" there is at least one result.
                        const results = (!data.length && affectedRows ? [[]] : data);
                        return results.map((_, idx) => parseRowQueryResult(results[idx], affectedRows, commands[idx]));
                    }
                    catch (err) {
                        if (err.code === mmsqlErrors.CANCELED) {
                            err.sqlectronError = 'CANCELED_BY_USER';
                        }
                        throw err;
                    }
                });
            },
            cancel: () => {
                if (!queryRequest) {
                    throw new Error('Query not ready to be canceled');
                }
                queryRequest.cancel();
            },
        };
    }
    async executeQuery(queryText, connection) {
        const { data, result } = await this.driverExecuteQuery({
            query: queryText,
            multiple: true,
        }, connection);
        const commands = utils_1.identifyCommands(queryText).map((item) => item.type);
        // Executing only non select queries will not return results.
        // So we "fake" there is at least one result.
        const rowsAffected = result.rowsAffected.reduce((a, b) => a + b, 0);
        const results = !data.length && rowsAffected ? [[]] : data;
        return results.map((value, idx) => {
            return parseRowQueryResult(value, rowsAffected, commands[idx]);
        });
    }
    async driverExecuteSingleQuery(queryArgs, connection) {
        const result = await this.driverExecuteQuery(queryArgs, connection);
        return {
            request: result.request,
            result: result.result,
            data: result.data[0],
        };
    }
    async driverExecuteQuery(queryArgs, connection) {
        const runQuery = async (connection) => {
            const request = connection.request();
            if (queryArgs.multiple) {
                request.multiple = true;
            }
            const result = await request.query(queryArgs.query);
            return {
                request,
                result,
                data: request.multiple
                    ? result.recordsets
                    : [result.recordset],
            };
        };
        return connection ? runQuery(connection) : this.runWithConnection(runQuery);
    }
    async runWithConnection(run) {
        if (!this.conn.connection) {
            this.conn.connection = await new mssql_1.ConnectionPool(this.conn.dbConfig).connect();
        }
        return run(this.conn.connection);
    }
    wrapIdentifier(value) {
        return wrapIdentifier(value);
    }
}
exports.default = SqlServerAdapter;
function wrapIdentifier(value) {
    return value !== '*' ? `[${value.replace(/\[/g, '[')}]` : '*';
}
exports.wrapIdentifier = wrapIdentifier;
function parseRowQueryResult(
// eslint-disable-next-line @typescript-eslint/no-explicit-any
data, affectedRows, command) {
    // Fallback in case the identifier could not reconize the command
    const isSelect = !!(data.length || !affectedRows);
    return {
        command: command || (isSelect && 'SELECT'),
        rows: data,
        fields: Object.keys(data[0] || {}).map((name) => ({ name })),
        rowCount: data.length,
        affectedRows,
    };
}
