"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.setSelectLimit = exports.clearSelectLimit = exports.Database = void 0;
const tunnel_1 = __importDefault(require("./tunnel"));
const adapters_1 = require("./adapters");
const logger_1 = __importDefault(require("./logger"));
const logger = logger_1.default('db');
const DEFAULT_LIMIT = 1000;
let selectLimit = null;
class Database {
    constructor(server, database) {
        this.connecting = false;
        this.connection = null;
        this.server = server;
        this.database = database;
    }
    handleSSHError() {
        return new Promise((resolve, reject) => {
            if (!this.server.sshTunnel) {
                return resolve();
            }
            this.server.sshTunnel.on('success', resolve);
            this.server.sshTunnel.on('error', (error) => {
                logger().error('ssh error %j', error);
                reject(error);
            });
        });
    }
    checkIsConnected() {
        if (this.connecting || !this.connection) {
            throw new Error('There is no connection available.');
        }
    }
    async connect() {
        /* eslint no-param-reassign: 0 */
        if (this.connecting) {
            throw new Error('There is already a connection in progress for this database. Aborting this new request.');
        }
        try {
            this.connecting = true;
            // terminate any previous lost connection for this DB
            if (this.connection) {
                void this.connection.disconnect();
            }
            // reuse existing tunnel
            if (this.server.config.ssh && !this.server.sshTunnel) {
                logger().debug('creating ssh tunnel');
                this.server.sshTunnel = await tunnel_1.default(this.server.config);
                const { address, port } = this.server.sshTunnel.address();
                logger().debug('ssh forwarding through local connection %s:%d', address, port);
                this.server.config.localHost = address;
                this.server.config.localPort = port;
            }
            const adapter = adapters_1.adapterFactory(this.server.config.adapter, this.server, this);
            await Promise.all([adapter.connect(), this.handleSSHError()]);
            this.connection = adapter;
        }
        catch (err) {
            logger().error('Connection error %j', err);
            this.disconnect();
            throw err;
        }
        finally {
            this.connecting = false;
        }
    }
    disconnect() {
        this.connecting = false;
        if (this.connection) {
            void this.connection.disconnect();
            this.connection = null;
        }
        this.server.removeDatabase(this.database);
    }
    getVersion() {
        this.checkIsConnected();
        return this.connection.getVersion();
    }
    listDatabases(filter) {
        this.checkIsConnected();
        return this.connection.listDatabases(filter);
    }
    listSchemas(filter) {
        this.checkIsConnected();
        return this.connection.listSchemas(filter);
    }
    listTables(filter) {
        this.checkIsConnected();
        return this.connection.listTables(filter);
    }
    listViews(filter) {
        this.checkIsConnected();
        return this.connection.listViews(filter);
    }
    listRoutines(filter) {
        this.checkIsConnected();
        return this.connection.listRoutines(filter);
    }
    listTableColumns(table, schema) {
        this.checkIsConnected();
        return this.connection.listTableColumns(table, schema);
    }
    listTableTriggers(table, schema) {
        this.checkIsConnected();
        return this.connection.listTableTriggers(table, schema);
    }
    listTableIndexes(table, schema) {
        this.checkIsConnected();
        return this.connection.listTableIndexes(table, schema);
    }
    getTableReferences(table, schema) {
        this.checkIsConnected();
        return this.connection.getTableReferences(table, schema);
    }
    getTableKeys(table, schema) {
        this.checkIsConnected();
        return this.connection.getTableKeys(table, schema);
    }
    query(queryText) {
        this.checkIsConnected();
        return this.connection.query(queryText);
    }
    executeQuery(queryText) {
        this.checkIsConnected();
        return this.connection.executeQuery(queryText);
    }
    getQuerySelectTop(table, schema, limit) {
        this.checkIsConnected();
        let limitValue = limit;
        if (limit === undefined) {
            limitValue = selectLimit !== null ? selectLimit : DEFAULT_LIMIT;
        }
        return Promise.resolve(this.connection.getQuerySelectTop(table, limitValue, schema));
    }
    getTableCreateScript(table, schema) {
        this.checkIsConnected();
        return this.connection.getTableCreateScript(table, schema);
    }
    async getTableSelectScript(table, schema) {
        const columnNames = await this.getTableColumnNames(table, schema);
        const schemaSelection = this.resolveSchema(schema);
        return [
            `SELECT ${columnNames.map((name) => this.wrap(name)).join(', ')}`,
            `FROM ${schemaSelection}${this.wrap(table)};`,
        ].join(' ');
    }
    async getTableInsertScript(table, schema) {
        const columnNames = await this.getTableColumnNames(table, schema);
        const schemaSelection = this.resolveSchema(schema);
        return [
            `INSERT INTO ${schemaSelection}${this.wrap(table)}`,
            `(${columnNames.map((name) => this.wrap(name)).join(', ')})\n`,
            `VALUES (${columnNames.fill('?').join(', ')});`,
        ].join(' ');
    }
    async getTableUpdateScript(table, schema) {
        const columnNames = await this.getTableColumnNames(table, schema);
        const setColumnForm = columnNames.map((col) => `${this.wrap(col)}=?`).join(', ');
        const schemaSelection = this.resolveSchema(schema);
        return [
            `UPDATE ${schemaSelection}${this.wrap(table)}\n`,
            `SET ${setColumnForm}\n`,
            'WHERE <condition>;',
        ].join(' ');
    }
    getTableDeleteScript(table, schema) {
        const schemaSelection = this.resolveSchema(schema);
        return Promise.resolve([`DELETE FROM ${schemaSelection}${this.wrap(table)}`, 'WHERE <condition>;'].join(' '));
    }
    getViewCreateScript(view, schema) {
        this.checkIsConnected();
        return this.connection.getViewCreateScript(view, schema);
    }
    getRoutineCreateScript(routine, type, schema) {
        this.checkIsConnected();
        return this.connection.getRoutineCreateScript(routine, type, schema);
    }
    truncateAllTables(schema) {
        return this.connection.truncateAllTables(schema);
    }
    async getTableColumnNames(table, schema) {
        this.checkIsConnected();
        const columns = await this.connection.listTableColumns(table, schema);
        return columns.map((column) => column.columnName);
    }
    resolveSchema(schema) {
        return schema ? `${this.wrap(schema)}.` : '';
    }
    wrap(identifier) {
        return this.connection.wrapIdentifier(identifier);
    }
}
exports.Database = Database;
function clearSelectLimit() {
    selectLimit = null;
}
exports.clearSelectLimit = clearSelectLimit;
function setSelectLimit(limit) {
    selectLimit = limit;
}
exports.setSelectLimit = setSelectLimit;
