"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.appendSemiColon = exports.identifyCommands = exports.versionCompare = exports.createCancelablePromise = exports.resolveHomePathToAbsolute = exports.readFile = void 0;
const fs_1 = __importDefault(require("fs"));
const os_1 = require("os");
const path_1 = __importDefault(require("path"));
const sql_query_identifier_1 = require("sql-query-identifier");
const errors_1 = require("./errors");
function readFile(filename) {
    const filePath = resolveHomePathToAbsolute(filename);
    return new Promise((resolve, reject) => {
        fs_1.default.readFile(path_1.default.resolve(filePath), { encoding: 'utf-8' }, (err, data) => {
            if (err)
                return reject(err);
            resolve(data);
        });
    });
}
exports.readFile = readFile;
function resolveHomePathToAbsolute(filename) {
    if (!/^~\//.test(filename)) {
        return filename;
    }
    return path_1.default.join(os_1.homedir(), filename.substring(2));
}
exports.resolveHomePathToAbsolute = resolveHomePathToAbsolute;
function createCancelablePromise(timeIdle = 100) {
    let canceled = false;
    let discarded = false;
    const wait = (time) => new Promise((resolve) => setTimeout(resolve, time));
    return {
        async wait() {
            while (!canceled && !discarded) {
                // eslint-disable-next-line
                await wait(timeIdle);
            }
            if (canceled) {
                throw new errors_1.CanceledByUserError();
            }
        },
        cancel() {
            canceled = true;
        },
        discard() {
            discarded = true;
        },
    };
}
exports.createCancelablePromise = createCancelablePromise;
/**
 * Compares two version strings.
 *
 * For two version strings, this fucntion will return -1 if the first version is smaller
 * than the second version, 0 if they are equal, and 1 if the second version is smaller.
 * However, this function will only compare up-to the smallest part of the version string
 * defined between the two, such '8' and '8.0.2' will be considered equal.
 */
function versionCompare(a, b) {
    const fullA = a.split('.').map((val) => parseInt(val, 10));
    const fullB = b.split('.').map((val) => parseInt(val, 10));
    for (let i = 0; i < Math.min(fullA.length, fullB.length); i++) {
        if (fullA[i] > fullB[i]) {
            return 1;
        }
        else if (fullA[i] < fullB[i]) {
            return -1;
        }
    }
    return 0;
}
exports.versionCompare = versionCompare;
function identifyCommands(queryText) {
    try {
        return sql_query_identifier_1.identify(queryText);
    }
    catch (err) {
        return [];
    }
}
exports.identifyCommands = identifyCommands;
function appendSemiColon(query) {
    let result = query.trim();
    if (result[result.length - 1] !== ';') {
        result += ';';
    }
    return result;
}
exports.appendSemiColon = appendSemiColon;
